// Copyright 2013 Canonical Ltd.  This software is licensed under the
// GNU Lesser General Public License version 3 (see the file COPYING).

/*
This is an example on how the Azure Go library can be used to interact with
the Windows Azure Service.
Note that this is a provided only as an example and that real code should
probably do something more sensible with errors than ignoring them or panicking.
*/
package main

import (
    "flag"
    "fmt"
    "launchpad.net/gwacl"
    "math/rand"
    "os"
    "time"
)

var certFile string
var subscriptionID string
var verbose bool
var pause bool

func getParams() error {
    flag.StringVar(&certFile, "cert", "", "Name of your management certificate file (in PEM format).")
    flag.StringVar(&subscriptionID, "subscriptionid", "", "Your Azure subscription ID.")
    flag.BoolVar(&verbose, "verbose", false, "Print debugging output")
    flag.BoolVar(&pause, "pause", false, "Wait for user input after the VM is brought up (useful for further testing)")

    flag.Parse()

    if certFile == "" {
        return fmt.Errorf("No .pem certificate specified.  Use the -cert option.")
    }
    if subscriptionID == "" {
        return fmt.Errorf("No subscription ID specified.  Use the -subscriptionid option.")
    }
    return nil
}

func checkError(err error) {
    if err != nil {
        panic(err)
    }
}

// makeRandomIdentifier creates an arbitrary identifier of the given length,
// consisting of only ASCII digits and lower-case ASCII letters.
// The identifier will start with the given prefix.  The prefix must be no
// longer than the specified length, or there'll be trouble.
func makeRandomIdentifier(prefix string, length int) string {
    // Only digits and lower-case ASCII letters are accepted.
    const chars = "abcdefghijklmnopqrstuvwxyz0123456789"

    if len(prefix) > length {
        panic(fmt.Errorf("prefix '%s' is more than the requested %d characters long", prefix, length))
    }

    id := prefix
    for len(id) < length {
        id += string(chars[rand.Intn(len(chars))])
    }
    return id
}

func main() {
    rand.Seed(int64(time.Now().Nanosecond()))

    err := getParams()
    if err != nil {
        fmt.Println(err)
        os.Exit(1)
    }

    api, err := gwacl.NewManagementAPI(subscriptionID, certFile)
    gwacl.SetVerbose(verbose)
    checkError(err)

    // Exercise the API.

    // Exercise hosted services API.
    ExerciseHostedServicesAPI(api)

    fmt.Println("All done.")
}

func ExerciseHostedServicesAPI(api *gwacl.ManagementAPI) {
    var err error
    networkConfig, err := api.GetNetworkConfiguration()
    checkError(err)
    if networkConfig == nil {
        fmt.Println("No network configuration is set.")
    } else {
        xml, err := networkConfig.Serialize()
        checkError(err)
        fmt.Println(xml)
    }

    location := "West US"
    release := "13.04"

    affinityGroupName := gwacl.MakeRandomHostname("affinitygroup")
    fmt.Printf("Creating an affinity group...")
    cag := gwacl.NewCreateAffinityGroup(affinityGroupName, "affinity-label", "affinity-description", location)
    err = api.CreateAffinityGroup(&gwacl.CreateAffinityGroupRequest{
        CreateAffinityGroup: cag})
    checkError(err)
    fmt.Println("created %s", affinityGroupName)

    defer func() {
        fmt.Println("Deleting affinity group %s", affinityGroupName)
        api.DeleteAffinityGroup(&gwacl.DeleteAffinityGroupRequest{
            Name: affinityGroupName})
        fmt.Println("Done deleting affinity group %s", affinityGroupName)
    }()

    fmt.Printf("Getting OS Image for release '%s' and location '%s'...\n", release, location)
    images, err := api.ListOSImages()
    checkError(err)
    preciseImage, err := images.GetLatestUbuntuImage(release, location)
    checkError(err)
    sourceImageName := preciseImage.Name
    fmt.Printf("Got image named '%s'\n", sourceImageName)
    fmt.Println("Done getting OS Image\n")

    hostServiceName := gwacl.MakeRandomHostedServiceName("gwacl")
    fmt.Printf("Creating a hosted service: '%s'...\n", hostServiceName)
    createHostedService := gwacl.NewCreateHostedServiceWithLocation(hostServiceName, "testLabel", location)
    err = api.AddHostedService(createHostedService)
    checkError(err)
    fmt.Println("Done creating a hosted service\n")

    defer func() {
        fmt.Println("Destroying hosted service...")
        api.DestroyHostedService(&gwacl.DestroyHostedServiceRequest{
            ServiceName: hostServiceName})
        fmt.Println("Done destroying hosted service\n")
    }()

    fmt.Println("Listing hosted services...")
    hostedServices, err := api.ListHostedServices()
    checkError(err)
    fmt.Printf("Got %d hosted service(s)\n", len(hostedServices))
    if len(hostedServices) > 0 {
        hostedService := hostedServices[0]
        detailedHostedService, err := api.GetHostedServiceProperties(hostedService.ServiceName, true)
        checkError(err)
        fmt.Printf("Hosted service '%s' contains %d deployments\n", hostedService.ServiceName, len(detailedHostedService.Deployments))
        // Do the same again with ListAllDeployments.
        deployments, err := api.ListAllDeployments(&gwacl.ListAllDeploymentsRequest{ServiceName: hostedService.ServiceName})
        checkError(err)
        if len(deployments) != len(detailedHostedService.Deployments) {
            fmt.Printf(
                "Mismatch in reported deployments: %d != %d",
                len(deployments), len(detailedHostedService.Deployments))
        }
    }
    fmt.Println("Done listing hosted services\n")

    fmt.Println("Adding VM deployment...")
    hostname := gwacl.MakeRandomHostname("gwaclhost")
    // Random passwords are no use to man nor beast here if you want to
    // test with your instance, so we'll use a fixed one.  It's not really a
    // security hazard in such a short-lived private instance.
    password := "Ubuntu123"
    username := "ubuntu"
    vhdName := makeRandomIdentifier("gwacldisk", 16)
    linuxConfigurationSet := gwacl.NewLinuxProvisioningConfigurationSet(hostname, username, password, "TEST_USER_DATA", "false")
    inputendpoint := gwacl.InputEndpoint{LocalPort: 22, Name: "sshport", Port: 22, Protocol: "TCP"}
    networkConfigurationSet := gwacl.NewNetworkConfigurationSet([]gwacl.InputEndpoint{inputendpoint})

    storageAccount := makeRandomIdentifier("gwacl", 24)
    storageLabel := makeRandomIdentifier("gwacl", 64)
    fmt.Printf("Requesting storage account with name '%s' and label '%s'...\n", storageAccount, storageLabel)
    cssi := gwacl.NewCreateStorageServiceInputWithLocation(storageAccount, storageLabel, location, "false")
    err = api.AddStorageAccount(cssi)
    checkError(err)
    fmt.Println("Done requesting storage account\n")

    defer func() {
        fmt.Printf("Deleting storage account %s...\n", storageAccount)
        api.DeleteStorageAccount(storageAccount)
        fmt.Println("Done deleting storage account\n")
    }()

    mediaLink := fmt.Sprintf("http://%s.blob.core.windows.net/vhds/%s.vhd", storageAccount, vhdName)
    diskName := makeRandomIdentifier("gwacldisk", 16)
    diskLabel := makeRandomIdentifier("gwacl", 64)
    vhd := gwacl.NewOSVirtualHardDisk("", diskLabel, diskName, mediaLink, sourceImageName, "Linux")
    roleName := makeRandomIdentifier("gwaclrole", 16)
    role := gwacl.NewRole("ExtraSmall", roleName, []gwacl.ConfigurationSet{*linuxConfigurationSet, *networkConfigurationSet}, []gwacl.OSVirtualHardDisk{*vhd})
    machineName := makeRandomIdentifier("gwaclmachine", 20)
    deployment := gwacl.NewDeploymentForCreateVMDeployment(machineName, "Staging", machineName, []gwacl.Role{*role}, "")
    err = api.AddDeployment(deployment, hostServiceName)
    checkError(err)
    fmt.Println("Done adding VM deployment\n")

    fmt.Println("Starting VM...")
    err = api.StartRole(&gwacl.StartRoleRequest{hostServiceName, deployment.Name, role.RoleName})
    checkError(err)
    fmt.Println("Done starting VM\n")

    fmt.Println("Listing VM...")
    instances, err := api.ListInstances(&gwacl.ListInstancesRequest{hostServiceName})
    checkError(err)
    fmt.Printf("Got %d instance(s)\n", len(instances))
    fmt.Println("Done listing VM\n")

    fmt.Println("Getting deployment info...")
    request := &gwacl.GetDeploymentRequest{ServiceName: hostServiceName, DeploymentName: machineName}
    deploy, err := api.GetDeployment(request)
    checkError(err)
    fqdn, err := deploy.GetFQDN()
    checkError(err)
    fmt.Println("Got deployment info\n")

    fmt.Println("host:", fqdn)
    fmt.Println("username:", username)
    fmt.Println("password:", password)
    fmt.Println("")

    if pause {
        var wait string
        fmt.Println("Pausing so you can play with the newly-created VM")
        fmt.Println("To clear up, type something and press enter to continue")
        fmt.Scan(&wait)
    }
}
