// Copyright 2013 Canonical Ltd.  This software is licensed under the
// GNU Lesser General Public License version 3 (see the file COPYING).

package gwacl

import (
    "fmt"
    . "launchpad.net/gocheck"
    "net/http"
)

type managementAPISuite struct{}

var _ = Suite(&managementAPISuite{})

// makeNamedRoleInstances creates an array of RoleInstance objects, each with
// the respective given name.
func makeNamedRoleInstances(names ...string) []RoleInstance {
    instances := make([]RoleInstance, 0)
    for _, name := range names {
        instances = append(instances, RoleInstance{RoleName: name})
    }
    return instances
}

// makeNamedDeployments creates an array of Deployment objects, each with
// the respective given name.
func makeNamedDeployments(names ...string) []Deployment {
    deployments := make([]Deployment, 0)
    for _, name := range names {
        deployments = append(deployments, Deployment{Name: name})
    }
    return deployments
}

// makeHostedService creates a HostedService with the given deployments.
func makeHostedService(deployments []Deployment) HostedService {
    return HostedService{
        ServiceName: "S1",
        Deployments: deployments,
    }
}

// makeOKXMLResponse creates a DispatcherResponse with status code OK, and
// an XML-serialized version of the given object.
// The response is wrapped in a slice because that's slightly easier for
// the callers.
func makeOKXMLResponse(c *C, bodyObject AzureObject) []DispatcherResponse {
    body, err := bodyObject.Serialize()
    c.Assert(err, IsNil)
    return []DispatcherResponse{
        {
            response: &x509Response{
                StatusCode: http.StatusOK,
                Body:       []byte(body),
            },
        },
    }
}

// TestListInstances goes through the happy path for ListInstances.
func (suite *managementAPISuite) TestListInstances(c *C) {
    service := makeHostedService(
        []Deployment{
            {RoleInstanceList: makeNamedRoleInstances("one", "two")},
            {RoleInstanceList: makeNamedRoleInstances("three", "four")},
        })
    record := []*X509Request{}
    rigRecordingPreparedResponseDispatcher(&record, makeOKXMLResponse(c, service))

    // Exercise ListInstances.
    api := makeAPI(c)
    request := &ListInstancesRequest{ServiceName: service.ServiceName}
    instances, err := api.ListInstances(request)
    c.Assert(err, IsNil)

    // We get the expected instances back.
    c.Check(instances, DeepEquals, []RoleInstance{
        service.Deployments[0].RoleInstanceList[0],
        service.Deployments[0].RoleInstanceList[1],
        service.Deployments[1].RoleInstanceList[0],
        service.Deployments[1].RoleInstanceList[1],
    })

    // The only request is for the service's properties
    c.Assert(record, Not(HasLen), 0)
    expectedURL := fmt.Sprintf(
        "%ssubscriptionId/services/hostedservices/%s?embed-detail=true",
        AZURE_URL, service.ServiceName)
    c.Check(record[0], DeepEquals, &X509Request{
        APIVersion: baseAPIVersion,
        URL:        expectedURL,
        Method:     "GET",
    })
}

func (suite *managementAPISuite) TestListInstancesFailsGettingDetails(c *C) {
    rigPreparedResponseDispatcher(
        []DispatcherResponse{{response: &x509Response{StatusCode: http.StatusNotFound}}},
    )
    api := makeAPI(c)
    request := &ListInstancesRequest{ServiceName: "SomeService"}
    instances, err := api.ListInstances(request)
    c.Check(instances, IsNil)
    c.Assert(err, NotNil)
    c.Assert(err.Error(), Equals, "GET request failed (404: Not Found)")
}

// TestListAllDeployments goes through the happy path for ListAllDeployments.
func (suite *managementAPISuite) TestListAllDeployments(c *C) {
    service := makeHostedService(makeNamedDeployments("one", "two"))
    record := []*X509Request{}
    rigRecordingPreparedResponseDispatcher(&record, makeOKXMLResponse(c, service))

    // Exercise ListDeployments.
    api := makeAPI(c)
    request := &ListAllDeploymentsRequest{ServiceName: service.ServiceName}
    deployments, err := api.ListAllDeployments(request)
    c.Assert(err, IsNil)

    // We get the complete set of deployments back.
    c.Check(deployments, DeepEquals, service.Deployments)

    // Only one request to the API is made.
    c.Assert(record, HasLen, 1)
}

// TestListDeployments tests ListDeployments, including filtering by name.
func (suite *managementAPISuite) TestListDeployments(c *C) {
    service := makeHostedService(makeNamedDeployments("Arthur", "Bobby"))
    record := []*X509Request{}
    rigRecordingPreparedResponseDispatcher(&record, makeOKXMLResponse(c, service))

    // Exercise ListDeployments.
    api := makeAPI(c)
    request := &ListDeploymentsRequest{
        ServiceName:     service.ServiceName,
        DeploymentNames: []string{"Arthur"},
    }
    deployments, err := api.ListDeployments(request)
    c.Assert(err, IsNil)

    // Only the first deployment - named "Arthur" - is returned.
    c.Check(deployments, DeepEquals, []Deployment{service.Deployments[0]})
    // Only one request to the API is made.
    c.Assert(record, HasLen, 1)
}

func (suite *managementAPISuite) TestListDeploymentsWithoutNamesReturnsNothing(c *C) {
    service := makeHostedService(makeNamedDeployments("One", "Two"))
    record := []*X509Request{}
    rigRecordingPreparedResponseDispatcher(&record, makeOKXMLResponse(c, service))
    // Exercise ListDeployments.
    api := makeAPI(c)
    request := &ListDeploymentsRequest{
        ServiceName:     service.ServiceName,
        DeploymentNames: []string{},
    }
    deployments, err := api.ListDeployments(request)
    c.Assert(err, IsNil)

    // No deployments are returned.
    c.Check(deployments, HasLen, 0)
}

func makeHostedServiceDescriptor() *HostedServiceDescriptor {
    url := MakeRandomString(10)
    serviceName := MakeRandomString(10)
    return &HostedServiceDescriptor{ServiceName: serviceName, URL: url}
}

func (suite *managementAPISuite) TestListSpecificHostedServices(c *C) {
    service1 := makeHostedServiceDescriptor()
    service2 := makeHostedServiceDescriptor()
    list := HostedServiceDescriptorList{HostedServices: []HostedServiceDescriptor{*service1, *service2}}
    XML, err := list.Serialize()
    c.Assert(err, IsNil)
    fixedResponse := x509Response{
        StatusCode: http.StatusOK,
        Body:       []byte(XML),
    }
    rigFixedResponseDispatcher(&fixedResponse)
    recordedRequests := make([]*X509Request, 0)
    rigRecordingDispatcher(&recordedRequests)

    api := makeAPI(c)
    request := &ListSpecificHostedServicesRequest{
        ServiceNames: []string{service1.ServiceName},
    }
    descriptors, err := api.ListSpecificHostedServices(request)

    // Only the first service is returned.
    c.Check(descriptors, DeepEquals, []HostedServiceDescriptor{*service1})
    // Only one request to the API is made.
    c.Assert(recordedRequests, HasLen, 1)
}

func (suite *managementAPISuite) TestListPrefixedHostedServices(c *C) {
    prefix := "prefix"
    service1 := &HostedServiceDescriptor{ServiceName: prefix + "service1"}
    service2 := makeHostedServiceDescriptor()
    list := HostedServiceDescriptorList{HostedServices: []HostedServiceDescriptor{*service1, *service2}}
    XML, err := list.Serialize()
    c.Assert(err, IsNil)
    fixedResponse := x509Response{
        StatusCode: http.StatusOK,
        Body:       []byte(XML),
    }
    rigFixedResponseDispatcher(&fixedResponse)
    recordedRequests := make([]*X509Request, 0)
    rigRecordingDispatcher(&recordedRequests)

    api := makeAPI(c)
    request := &ListPrefixedHostedServicesRequest{
        ServiceNamePrefix: prefix,
    }
    descriptors, err := api.ListPrefixedHostedServices(request)

    // Only the first service is returned.
    c.Check(descriptors, DeepEquals, []HostedServiceDescriptor{*service1})
    // Only one request to the API is made.
    c.Assert(recordedRequests, HasLen, 1)
}

func (suite *managementAPISuite) TestListSpecificHostedServicesWithoutNamesReturnsNothing(c *C) {
    service1 := makeHostedServiceDescriptor()
    service2 := makeHostedServiceDescriptor()
    list := HostedServiceDescriptorList{HostedServices: []HostedServiceDescriptor{*service1, *service2}}
    XML, err := list.Serialize()
    c.Assert(err, IsNil)
    fixedResponse := x509Response{
        StatusCode: http.StatusOK,
        Body:       []byte(XML),
    }
    rigFixedResponseDispatcher(&fixedResponse)
    recordedRequests := make([]*X509Request, 0)
    rigRecordingDispatcher(&recordedRequests)

    api := makeAPI(c)
    request := &ListSpecificHostedServicesRequest{
        ServiceNames: []string{},
    }
    descriptors, err := api.ListSpecificHostedServices(request)

    c.Check(descriptors, DeepEquals, []HostedServiceDescriptor{})
    // Only one request to the API is made.
    c.Assert(recordedRequests, HasLen, 1)
}

var exampleOkayResponse = DispatcherResponse{
    response: &x509Response{StatusCode: http.StatusOK},
}

var exampleFailResponse = DispatcherResponse{
    response: &x509Response{StatusCode: http.StatusInternalServerError},
}

var exampleNotFoundResponse = DispatcherResponse{
    response: &x509Response{StatusCode: http.StatusNotFound},
}

type suiteDestroyDeployment struct{}

var _ = Suite(&suiteDestroyDeployment{})

func (suite *suiteDestroyDeployment) makeExampleDeployment() *Deployment {
    return &Deployment{
        RoleInstanceList: makeNamedRoleInstances("one", "two"),
        RoleList: []Role{
            {OSVirtualHardDisk: []OSVirtualHardDisk{
                {DiskName: "disk1"}, {DiskName: "disk2"}}},
            {OSVirtualHardDisk: []OSVirtualHardDisk{
                {DiskName: "disk1"}, {DiskName: "disk3"}}},
        },
    }
}

func (suite *suiteDestroyDeployment) TestHappyPath(c *C) {
    var responses []DispatcherResponse
    // Prepare.
    exampleDeployment := suite.makeExampleDeployment()
    responses = append(responses, makeOKXMLResponse(c, exampleDeployment)...)
    // For deleting the deployment.
    responses = append(responses, exampleOkayResponse)
    // For deleting disks.
    responses = append(responses, exampleOkayResponse, exampleOkayResponse, exampleOkayResponse)
    record := []*X509Request{}
    rigRecordingPreparedResponseDispatcher(&record, responses)
    // Exercise DestroyDeployment.
    api := makeAPI(c)
    request := &DestroyDeploymentRequest{
        ServiceName:    "service-name",
        DeploymentName: "deployment-name",
    }
    err := api.DestroyDeployment(request)
    c.Assert(err, IsNil)
    c.Check(record, HasLen, 5)
    assertGetDeploymentRequest(c, api, &GetDeploymentRequest{
        request.ServiceName, request.DeploymentName}, record[0])
    assertDeleteDeploymentRequest(c, api, request.ServiceName,
        request.DeploymentName, record[1])
    assertDeleteDiskRequest(c, api, "disk1", record[2])
    assertDeleteDiskRequest(c, api, "disk2", record[3])
    assertDeleteDiskRequest(c, api, "disk3", record[4])
}

func (suite *suiteDestroyDeployment) TestOkayWhenDeploymentNotFound(c *C) {
    var responses []DispatcherResponse
    // Prepare.
    responses = []DispatcherResponse{exampleNotFoundResponse}
    record := []*X509Request{}
    rigRecordingPreparedResponseDispatcher(&record, responses)
    // Exercise DestroyDeployment.
    api := makeAPI(c)
    request := &DestroyDeploymentRequest{
        ServiceName:    "service-name",
        DeploymentName: "deployment-name",
    }
    err := api.DestroyDeployment(request)
    c.Assert(err, IsNil)
    c.Check(record, HasLen, 1)
}

func (suite *suiteDestroyDeployment) TestOkayWhenAssetsNotFound(c *C) {
    var responses []DispatcherResponse
    // Prepare.
    exampleDeployment := suite.makeExampleDeployment()
    responses = append(responses, makeOKXMLResponse(c, exampleDeployment)...)
    // For deleting the deployment.
    responses = append(responses, exampleNotFoundResponse)
    // For deleting the disks.
    responses = append(responses, exampleNotFoundResponse, exampleNotFoundResponse, exampleNotFoundResponse)
    record := []*X509Request{}
    rigRecordingPreparedResponseDispatcher(&record, responses)
    // Exercise DestroyDeployment.
    api := makeAPI(c)
    request := &DestroyDeploymentRequest{
        ServiceName:    "service-name",
        DeploymentName: "deployment-name",
    }
    err := api.DestroyDeployment(request)
    c.Assert(err, IsNil)
    c.Check(record, HasLen, 5)
    assertGetDeploymentRequest(c, api, &GetDeploymentRequest{
        request.ServiceName, request.DeploymentName}, record[0])
    assertDeleteDeploymentRequest(c, api, request.ServiceName,
        request.DeploymentName, record[1])
    assertDeleteDiskRequest(c, api, "disk1", record[2])
    assertDeleteDiskRequest(c, api, "disk2", record[3])
    assertDeleteDiskRequest(c, api, "disk3", record[4])
}

func (suite *suiteDestroyDeployment) TestFailsGettingDeployment(c *C) {
    var responses []DispatcherResponse
    // Prepare.
    responses = []DispatcherResponse{exampleFailResponse}
    record := []*X509Request{}
    rigRecordingPreparedResponseDispatcher(&record, responses)
    // Exercise DestroyDeployment.
    api := makeAPI(c)
    request := &DestroyDeploymentRequest{
        ServiceName:    "service-name",
        DeploymentName: "deployment-name",
    }
    err := api.DestroyDeployment(request)
    c.Assert(err, NotNil)
    c.Check(err, ErrorMatches, "GET request failed [(]500: Internal Server Error[)]")
    c.Check(record, HasLen, 1)
}

func (suite *suiteDestroyDeployment) TestFailsDeletingDisk(c *C) {
    var responses []DispatcherResponse
    // Prepare.
    exampleDeployment := suite.makeExampleDeployment()
    responses = append(responses, makeOKXMLResponse(c, exampleDeployment)...)
    // For deleting disks.
    responses = append(responses, exampleFailResponse)
    record := []*X509Request{}
    rigRecordingPreparedResponseDispatcher(&record, responses)
    // Exercise DestroyDeployment.
    api := makeAPI(c)
    request := &DestroyDeploymentRequest{
        ServiceName:    "service-name",
        DeploymentName: "deployment-name",
    }
    err := api.DestroyDeployment(request)
    c.Assert(err, NotNil)
    c.Check(err, ErrorMatches, "DELETE request failed [(]500: Internal Server Error[)]")
    c.Check(record, HasLen, 2)
}

func (suite *suiteDestroyDeployment) TestFailsDeletingDeployment(c *C) {
    var responses []DispatcherResponse
    // Prepare.
    exampleDeployment := suite.makeExampleDeployment()
    responses = append(responses, makeOKXMLResponse(c, exampleDeployment)...)
    // For deleting disks.
    responses = append(responses, exampleOkayResponse, exampleOkayResponse, exampleOkayResponse)
    // For other requests.
    responses = append(responses, exampleFailResponse)
    record := []*X509Request{}
    rigRecordingPreparedResponseDispatcher(&record, responses)
    // Exercise DestroyDeployment.
    api := makeAPI(c)
    request := &DestroyDeploymentRequest{
        ServiceName:    "service-name",
        DeploymentName: "deployment-name",
    }
    err := api.DestroyDeployment(request)
    c.Assert(err, NotNil)
    c.Check(err, ErrorMatches, "DELETE request failed [(]500: Internal Server Error[)]")
    c.Check(record, HasLen, 5)
}

type suiteDestroyHostedService struct{}

var _ = Suite(&suiteDestroyHostedService{})

func (suite *suiteDestroyHostedService) makeExampleHostedService(deploymentNames ...string) *HostedService {
    var exampleHostedService = &HostedService{}
    for _, deploymentName := range deploymentNames {
        exampleHostedService.Deployments = append(
            exampleHostedService.Deployments,
            Deployment{Name: deploymentName})
    }
    return exampleHostedService
}

func (suite *suiteDestroyHostedService) TestHappyPath(c *C) {
    var responses []DispatcherResponse
    // DestroyHostedService first gets the hosted service proerties.
    exampleHostedService := suite.makeExampleHostedService("one", "two")
    responses = append(responses, makeOKXMLResponse(c, exampleHostedService)...)
    // It calls DestroyDeployment, which first gets the deployment's
    // properties, deletes any assets contained therein (none in this case)
    // then deletes the deployment.
    for _, deployment := range exampleHostedService.Deployments {
        responses = append(responses, makeOKXMLResponse(c, &deployment)...)
        responses = append(responses, exampleOkayResponse)
    }
    // For deleting the service itself.
    responses = append(responses, exampleOkayResponse)
    record := []*X509Request{}
    rigRecordingPreparedResponseDispatcher(&record, responses)
    // Exercise DestroyHostedService.
    api := makeAPI(c)
    request := &DestroyHostedServiceRequest{
        ServiceName: "service-name",
    }
    err := api.DestroyHostedService(request)
    c.Assert(err, IsNil)
    c.Check(record, HasLen, 6)
    // The first request is for the hosted service.
    assertGetHostedServicePropertiesRequest(c, api, request.ServiceName, true, record[0])
    // The second and third requests fetch then delete deployment "one"; see
    // DestroyDeployment for an explanation of this behaviour.
    assertGetDeploymentRequest(c, api, &GetDeploymentRequest{
        request.ServiceName, "one"}, record[1])
    assertDeleteDeploymentRequest(c, api, request.ServiceName, "one",
        record[2])
    // The fourth and fifth requests are a repaat of the previous two, but for
    // deployment "two".
    assertGetDeploymentRequest(c, api, &GetDeploymentRequest{
        request.ServiceName, "two"}, record[3])
    assertDeleteDeploymentRequest(c, api, request.ServiceName, "two",
        record[4])
    // The last request deletes the hosted service.
    assertDeleteHostedServiceRequest(c, api, request.ServiceName, record[5])
}

func (suite *suiteDestroyHostedService) TestOkayWhenHostedServiceNotFound(c *C) {
    var responses []DispatcherResponse
    // Prepare.
    responses = []DispatcherResponse{exampleNotFoundResponse}
    record := []*X509Request{}
    rigRecordingPreparedResponseDispatcher(&record, responses)
    // Exercise DestroyHostedService.
    api := makeAPI(c)
    request := &DestroyHostedServiceRequest{ServiceName: "service-name"}
    err := api.DestroyHostedService(request)
    c.Assert(err, IsNil)
    c.Check(record, HasLen, 1)
}

func (suite *suiteDestroyHostedService) TestOkayWhenDeploymentsNotFound(c *C) {
    var responses []DispatcherResponse
    // Prepare.
    exampleHostedService := suite.makeExampleHostedService("one", "two")
    responses = append(responses, makeOKXMLResponse(c, exampleHostedService)...)
    // Someone else has destroyed the deployments in the meantime.
    responses = append(responses, exampleNotFoundResponse, exampleNotFoundResponse)
    // Success deleting the hosted service.
    responses = append(responses, exampleOkayResponse)
    record := []*X509Request{}
    rigRecordingPreparedResponseDispatcher(&record, responses)
    // Exercise DestroyHostedService.
    api := makeAPI(c)
    request := &DestroyHostedServiceRequest{ServiceName: "service-name"}
    err := api.DestroyHostedService(request)
    c.Assert(err, IsNil)
    c.Check(record, HasLen, 4)
    assertGetDeploymentRequest(c, api, &GetDeploymentRequest{
        request.ServiceName, "one"}, record[1])
    assertGetDeploymentRequest(c, api, &GetDeploymentRequest{
        request.ServiceName, "two"}, record[2])
    assertDeleteHostedServiceRequest(c, api, request.ServiceName, record[3])
}

func (suite *suiteDestroyHostedService) TestOkayWhenHostedServiceNotFoundWhenDeleting(c *C) {
    var responses []DispatcherResponse
    // Prepare.
    exampleHostedService := suite.makeExampleHostedService("one", "two")
    responses = append(responses, makeOKXMLResponse(c, exampleHostedService)...)
    // Someone else has destroyed the deployments in the meantime.
    responses = append(responses, exampleNotFoundResponse, exampleNotFoundResponse)
    // Someone else has destroyed the hosted service in the meantime.
    responses = append(responses, exampleNotFoundResponse)
    record := []*X509Request{}
    rigRecordingPreparedResponseDispatcher(&record, responses)
    // Exercise DestroyHostedService.
    api := makeAPI(c)
    request := &DestroyHostedServiceRequest{ServiceName: "service-name"}
    err := api.DestroyHostedService(request)
    c.Assert(err, IsNil)
    c.Check(record, HasLen, 4)
    assertGetDeploymentRequest(c, api, &GetDeploymentRequest{
        request.ServiceName, "one"}, record[1])
    assertGetDeploymentRequest(c, api, &GetDeploymentRequest{
        request.ServiceName, "two"}, record[2])
    assertDeleteHostedServiceRequest(c, api, request.ServiceName, record[3])
}

func (suite *suiteDestroyHostedService) TestFailsGettingHostedService(c *C) {
    var responses []DispatcherResponse
    // Prepare.
    responses = []DispatcherResponse{exampleFailResponse}
    record := []*X509Request{}
    rigRecordingPreparedResponseDispatcher(&record, responses)
    // Exercise DestroyHostedService.
    api := makeAPI(c)
    request := &DestroyHostedServiceRequest{ServiceName: "service-name"}
    err := api.DestroyHostedService(request)
    c.Assert(err, NotNil)
    c.Check(err, ErrorMatches, "GET request failed [(]500: Internal Server Error[)]")
    c.Check(record, HasLen, 1)
}

func (suite *suiteDestroyHostedService) TestFailsDestroyingDeployment(c *C) {
    var responses []DispatcherResponse
    // Prepare.
    exampleHostedService := suite.makeExampleHostedService("one", "two")
    responses = append(responses, makeOKXMLResponse(c, exampleHostedService)...)
    responses = append(responses, exampleFailResponse)
    record := []*X509Request{}
    rigRecordingPreparedResponseDispatcher(&record, responses)
    // Exercise DestroyHostedService.
    api := makeAPI(c)
    request := &DestroyHostedServiceRequest{ServiceName: "service-name"}
    err := api.DestroyHostedService(request)
    c.Assert(err, NotNil)
    c.Check(err, ErrorMatches, "GET request failed [(]500: Internal Server Error[)]")
    c.Check(record, HasLen, 2)
}

func (suite *suiteDestroyHostedService) TestFailsDeletingHostedService(c *C) {
    var responses []DispatcherResponse
    // Prepare.
    exampleHostedService := suite.makeExampleHostedService("one", "two")
    responses = append(responses, makeOKXMLResponse(c, exampleHostedService)...)
    // Deployments not found, but that's okay.
    responses = append(responses, exampleNotFoundResponse, exampleNotFoundResponse)
    // When deleting hosted service.
    responses = append(responses, exampleFailResponse)
    record := []*X509Request{}
    rigRecordingPreparedResponseDispatcher(&record, responses)
    // Exercise DestroyHostedService.
    api := makeAPI(c)
    request := &DestroyHostedServiceRequest{ServiceName: "service-name"}
    err := api.DestroyHostedService(request)
    c.Assert(err, NotNil)
    c.Check(err, ErrorMatches, "DELETE request failed [(]500: Internal Server Error[)]")
    c.Check(record, HasLen, 4)
}
